from datetime import datetime
import json
from json.decoder import JSONDecodeError
from pathlib import Path
from typing import Dict, Generator, Iterator, List, Optional, Tuple, Union

from geomet import wkt
import requests
from requests import Response, Session

from .types import DATETIME_OR_STR_TUPLE, SEARCH_PREDICATE_CHOICE

DEFAULT_RGD_API = 'https://www.resonantgeodata.com/api'

API_KEY_DIR_PATH = Path('~/.rgd/').expanduser()
API_KEY_FILE_NAME = 'token'


def pager(session: Session, url: str, **kwargs) -> Iterator[Response]:
    """Exhaust a DRF Paginated list."""
    while True:
        r = session.get(url, **kwargs)
        yield r

        if 'next' in r.links:
            url = r.links['next']['url']
        else:
            break


def limit_offset_pager(session: Session, url: str, **kwargs) -> Generator[Dict, None, None]:
    """Exhaust a DRF Paginated list, respecting limit/offset."""
    # Default params kwarg
    if not kwargs.get('params'):
        kwargs['params'] = {}

    params: Optional[Dict] = kwargs['params']
    total_limit = params.get('limit')

    # Default offset
    if params.get('offset') is None:
        params['offset'] = 0

    num_results = 0
    while True:
        # Update limit
        if total_limit:
            params['limit'] = total_limit - num_results

        # Make request and raise exception if failed
        r = session.get(url, **kwargs)
        r.raise_for_status()

        # Yield results
        results = r.json()['results']
        yield from results

        # Update offset and num_results
        params['offset'] += len(results)
        num_results += len(results)

        # Check if there is no more data, or if we've reached our limit
        no_more_data = 'next' not in r.links or not results
        limit_reached = total_limit and num_results >= total_limit
        if no_more_data or limit_reached:
            break


def iterate_response_bytes(
    url: str, chunk_size: int = 1024 * 1024, raise_for_status: bool = True
) -> Iterator[bytes]:
    """Return the response body as an iterator of bytes, where each item is `chunk_size` bytes long."""
    r = requests.get(url, stream=True)

    if raise_for_status:
        r.raise_for_status()

    return r.iter_content(chunk_size=chunk_size)


def datetime_to_str(value: object):
    """Convert datetime objects to ISO8601 strings."""
    if value is not None:
        if isinstance(value, datetime):
            return value.isoformat()

    return value


def datetime_to_time(value: object):
    """Convert datetime objects to HH:MM strings."""
    if value is not None:
        if isinstance(value, datetime):
            return value.strftime('%H:%M')

    return value


def order_datetimes(value1: object, value2: object):
    """
    Sort 2 objects if they are datetimes.

    Example:
        >>> from rgd_client.utils import *  # NOQA
        >>> value1 = datetime.fromisoformat('2000-01-01T00:13:30')
        >>> value2 = datetime.fromisoformat('1999-01-01T00:13:30')
        >>> result = order_datetimes(value1, value2)
        >>> assert len(result) == 2
    """
    if isinstance(value1, datetime) and isinstance(value2, datetime):
        return (value1, value2) if value1 < value2 else (value2, value1)

    return value1, value2


def spatial_subentry_id(search_result):
    """Get the id of a returned SpatialEntry."""
    if 'stac_version' in search_result:
        return search_result['id']
    return search_result['spatial_id']


def spatial_search_params(
    query: Optional[Union[Dict, str]] = None,
    predicate: Optional[SEARCH_PREDICATE_CHOICE] = None,
    relates: Optional[str] = None,
    distance: Optional[Tuple[float, float]] = None,
    acquired: Optional[DATETIME_OR_STR_TUPLE] = None,
    instrumentation: Optional[str] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
    time_of_day: Optional[DATETIME_OR_STR_TUPLE] = None,
    collections: Optional[List[int]] = None,
) -> Dict:
    # The dict that will be used to store params.
    # Initialize with queries that won't be additionally processed.
    params = {
        'predicate': predicate,
        'relates': relates,
        'instrumentation': instrumentation,
        'limit': limit,
        'offset': offset,
    }

    if query:
        if isinstance(query, str):
            try:
                query = json.loads(query)
            except JSONDecodeError:
                pass

        if isinstance(query, dict):
            # Allow failure on invalid format
            query = wkt.dumps(query)

        params['q'] = query

    # Process range params

    if distance and len(distance) == 2:
        dmin, dmax = distance
        params['distance_min'] = dmin
        params['distance_max'] = dmax

    # TODO: Determine if the before/after param order needs to be swapped?
    if acquired and len(acquired) == 2:
        amin, amax = order_datetimes(*acquired)
        params['acquired_before'] = datetime_to_str(amax)
        params['acquired_after'] = datetime_to_str(amin)

    if time_of_day and len(time_of_day) == 2:
        after, before = order_datetimes(*time_of_day)
        params['time_of_day_after'] = datetime_to_time(after)
        params['time_of_day_before'] = datetime_to_time(before)

    if collections:
        params['collections'] = collections

    return params


def download_checksum_file_to_path(
    file: Dict,
    path: Optional[Path] = None,
    keep_existing=False,
    use_id=False,
) -> Path:
    """
    Download a RGD ChecksumFile to a given path.

    Args:
        file: A RGD ChecksumFile serialized as a Dict.
        path: The root path to download this file to. Defaults to the current working directory.
        keep_existing: If False, replace files existing on disk.
        use_id: If True, save this file to disk using it's ID, rather than it's name.

    Returns:
        The path on disk the file was downloaded to.
    """
    filepath: str = str(file['id']) if use_id else file['name']
    file_download_url: str = file['download_url']

    if not path:
        path = Path.cwd()

    # Parse file path to identifiy nested directories
    filepath: str = filepath.lstrip('/')
    split_filepath: List[str] = filepath.split('/')
    parent_dirname = '/'.join(split_filepath[:-1])
    filename = split_filepath[-1]

    # Create nested directory if necessary
    parent_path = path / parent_dirname if parent_dirname else path
    parent_path.mkdir(parents=True, exist_ok=True)

    # Download contents to file
    file_path = parent_path / filename
    if not (file_path.is_file() and keep_existing):
        with open(file_path, 'wb') as open_file_path:
            for chunk in iterate_response_bytes(file_download_url):
                open_file_path.write(chunk)

    return file_path



from typing import Callable, Union

from rgd_client.client import RgdClient
from torch.utils.data import Dataset


class RemoteDataset(Dataset):
    """
    Remote dataset interface for PyTorch.

    Parameters
    ----------
    collection : int, str
        ID of the Collection to source data from (can also be the name of the
        collection).
    client: RgdClient
        The client connected to an RGD instance.
    data_handler : Callable
        A callable that will handle converting the data from a ChecksumFile to
        something usable by PyTorch.

    """

    def __init__(
        self, collection: Union[int, str, dict], client: RgdClient, data_handler: Callable = None
    ):
        self._client = client
        self._collection_identifier = collection
        self._collection = None
        self._callback = data_handler
        # Prefetch collection to make sure valid
        self.collection

    @property
    def collection(self):
        if self._collection is None:
            if isinstance(self._collection_identifier, (int, dict)):
                collection = self._client.rgd.get_collection(self._collection_identifier)
            elif isinstance(self._collection_identifier, str):
                collection = self._client.rgd.get_collection_by_name(self._collection_identifier)
            else:
                raise ValueError(f'Collection {self._collection_identifier} not understood.')
            self._collection = collection
        return self._collection

    def __getitem__(self, index: str):
        item = self._client.rgd.get_collection_item(self.collection, index)
        if self._callback is not None:
            return self._callback(item)
        return item

    def __len__(self):
        return self.collection['len']

    def reset(self):
        """Clear the cached collection metadata."""
        self._collection = None



import getpass
import logging
import os
from typing import List, Optional, Type

import requests

from .plugin import CorePlugin
from .session import RgdClientSession, clone_session
from .utils import API_KEY_DIR_PATH, API_KEY_FILE_NAME, DEFAULT_RGD_API

logger = logging.getLogger(__name__)


class RgdClient:
    def __init__(
        self,
        api_url: str = DEFAULT_RGD_API,
        username: Optional[str] = None,
        password: Optional[str] = None,
        save: Optional[bool] = True,
    ) -> None:
        """
        Initialize the base RGD Client.

        Args:
            api_url: The base url of the RGD API instance.
            username: The username to authenticate to the instance with, if any.
            password: The password associated with the provided username. If None, a prompt will be provided.
            save: Whether or not to save the logged-in user's API key to disk for future use.

        Returns:
            A base RgdClient instance.
        """
        # Look for an API key in the environment. If it's not there, check username/password
        api_key = _read_api_key(api_url=api_url, username=username, password=password)
        if api_key is None:
            if username is not None and password is None:
                password = getpass.getpass()

            # Get an API key for this user and save it to disk
            if username and password:
                api_key = _get_api_key(api_url, username, password, save)
                if api_key is None:
                    logger.error(
                        'Failed to retrieve API key; are your username and password correct?'
                    )

        self.session = RgdClientSession(base_url=api_url, auth_token=api_key)
        self.rgd = CorePlugin(clone_session(self.session))

    def clear_token(self):
        """Delete a locally-stored API key."""
        (API_KEY_DIR_PATH / API_KEY_FILE_NAME).unlink(missing_ok=True)


def _get_api_key(api_url: str, username: str, password: str, save: bool) -> Optional[str]:
    """Get an RGD API Key for the given user from the server, and save it if requested."""
    resp = requests.post(f'{api_url}/api-token-auth', {'username': username, 'password': password})
    token = resp.json().get('token')
    if token is None:
        return None
    if save:
        API_KEY_DIR_PATH.mkdir(parents=True, exist_ok=True)
        with open(API_KEY_DIR_PATH / API_KEY_FILE_NAME, 'w') as fd:
            fd.write(token)
    return token


def _read_api_key(api_url: str, username: str = None, password: str = None) -> Optional[str]:
    """
    Retrieve an RGD API Key from the users environment.

    This function checks for an environment variable named RGD_API_TOKEN and returns it if it exists.
    If it does not exist, it looks for a file located at ~/.rgd/token and returns its contents.
    """
    api_key = os.getenv('RGD_API_TOKEN', None)
    save = False
    if api_key is None:
        try:
            # read the first line of the text file at ~/.rgd/token
            with open(API_KEY_DIR_PATH / API_KEY_FILE_NAME, 'r') as fd:
                api_key = fd.readline().strip()
                save = True  # save any new api key to disk
        except FileNotFoundError:
            return None

    # Make sure API key works by hitting a protected endpoint
    resp = requests.get(f'{api_url}/rgd/collection', headers={'Authorization': f'Token {api_key}'})

    # If it doesn't, try to get a new one and save it to ~/.rgd/token, as the current one is corrupted
    if resp.status_code == 401:
        logger.error('API key is invalid.')
        # If username + password were provided, try to get a new API key with them
        # Note we only do this if `save` is `True`, i.e. if the user originally attempted to
        # instantiate the client with an API key located on disk. If they instead provided an env
        # var, do not assume that they want a key saved and do not attempt to fetch a new one.
        if save and username is not None and password is not None:
            logger.info('Attempting to fetch a new API key...')
            api_key = _get_api_key(api_url, username, password, save=True)
            if api_key is not None:
                logger.info('Succeeded.')
            return api_key
        else:
            logger.error('Provide your username and password next time to fetch a new one.')
            return None

    # If the response failed with an error status other than 401, raise an exception
    resp.raise_for_status()

    return api_key


def create_rgd_client(
    api_url: str = DEFAULT_RGD_API,
    username: Optional[str] = None,
    password: Optional[str] = None,
    save: Optional[bool] = True,
    extra_plugins: Optional[List[Type]] = None,
):
    # Avoid circular import
    from ._plugin_utils import _inject_plugin_deps, _plugin_classes, _plugin_instances

    # Strip trailing slash
    if api_url.endswith('/'):
        api_url = api_url.rstrip('/')

    # Create initial client
    client = RgdClient(api_url, username, password, save)

    # Perform plugin initialization
    plugin_classes = _plugin_classes(extra_plugins=extra_plugins)
    plugin_instances = _plugin_instances(client, plugin_classes)
    _inject_plugin_deps(plugin_instances)

    return client
